from iclik.inform_crit import claic, clbic
import numpy as np
import math

# case 1
def f1(params):
    x, y, z = params
    return x**2 + y**2 + z**2

params1 = (1,2,3)
wolfram_hessmat1 = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]])
wolfram_jacmat1 = np.array([[2*1, 2*2, 2*3]])

def test_claic():
    likelihood = f1(params1)
    claic_manual = (-2*likelihood + 2 * np.trace(wolfram_jacmat1 
                               @ np.linalg.inv(wolfram_hessmat1)))
    
    assert math.isclose(claic_manual, claic(f1, params1), abs_tol=1e-5)

def test_clbic():
    likelihood = f1(params1)
    n_samples = 100
    clbic_manual = (-2 * likelihood
                    + np.log(n_samples)
                    * np.trace(wolfram_jacmat1 
                               @ np.linalg.inv(wolfram_hessmat1)))
    
    assert math.isclose(clbic_manual,
                                       clbic(f1, params1, n_samples=100), abs_tol=1e-5)






import numpy as np
from iclik.matrices import hessian, jacobian

# case 1
def f1(params):
    x, y, z = params
    return x**2 + 2*y + z**2

params1 = (5,6,7)
wolfram_hessmat1 = np.array([[2, 0, 0], [0, 0, 0], [0, 0, 2]])
wolfram_jacmat1 = np.array([[2*5, 2, 2*7]])

def test_hessian_case1():
    hessmat = hessian(f1, params1, step=1e-2, step_factor=0.1)
    np.testing.assert_allclose(hessmat, wolfram_hessmat1, atol=1e-5)

def test_jacobian_case1():
    jacmat = jacobian(f1, params1, step=1e-2, step_factor=0.1)
    np.testing.assert_allclose(jacmat, wolfram_jacmat1, atol=1e-5)



import numpy as np
import numdifftools as nd


def hessian(f, estimates, step, step_factor):
    """Calculate Hessian by iteratively changing step size until all elements are finite.

    Args:
        f (function): Composite log likelihood function.
        estimates (ndarray): Parameter estimates at which to evaluate.
        initial_step (float): Initial step size for calculations.
        step_factor (float): Factor to multiply step size by at each iteration.

    Returns:
        ndarray: Hessian matrix.
    """
    hess_f = nd.core.Hessian(f, step=step)
    hess_mat = hess_f(np.array(estimates))
    if not (np.isfinite(hess_mat)).all():
            step = step * step_factor
            hess_mat = hessian(f, estimates, step=step, step_factor=step_factor)

    return hess_mat


def jacobian(f, estimates, step, step_factor):
    """Calculate Jacobian by iteratively changing step size until all elements are finite.

    Args:
        f (function): Composite log likelihood function.
        estimates (ndarray): Parameter estimates at which to evaluate.
        initial_step (float): Initial step size for calculations.
        step_factor (float): Factor to multiply step size by at each iteration.

    Returns:
        ndarray: Jacobian matrix.
    """
    jac_f = nd.core.Jacobian(f, step=step)
    jac_mat = jac_f(np.array(estimates))
    if not (np.isfinite(jac_mat)).all():
            step = step * step_factor
            jac_mat = jacobian(f, estimates, step=step, step_factor=step_factor)

    return jac_mat


from setuptools import setup

setup(
    name='iclik',
    version='0.1.10-alpha',
    packages=['iclik'],
    description="Information criteria for composite likelihood models",
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    install_requires=[
        "numdifftools",
        "numpy",
        "scipy"
    ]
)



import numpy as np
from iclik.matrices import jacobian, hessian
import warnings


def _make_matrices(f, estimates, step, step_factor):
     jac = jacobian(f, estimates, step, step_factor)
     hess = hessian(f, estimates, step, step_factor)

     return jac, hess


def _cl_akaike(likelihood, jacobian_mat, hessian_mat):
    """Compute CLAIC from matrices."""
    return (-2 * likelihood 
            + 2 * np.trace(jacobian_mat @ np.linalg.inv(hessian_mat)))


def _cl_bayes(likelihood, jacobian_mat, hessian_mat, n_samples):
    """Compute CLBIC from matrices."""
    return (-2 * likelihood
            + np.log(n_samples)
            * np.trace(jacobian_mat @ np.linalg.inv(hessian_mat)))


def _detect_zeros(estimates, response="warn"):
    """Detect whether array of parameter estimates contains zeros."""
    if (estimates <= 0).any():
        if response=="error":
            raise ValueError("Zero or negative values detected")
        elif response=="warn":
            warnings.warn("Zeros or negative values detected in input - information criterion cannot be calculated", UserWarning)
        else:
             raise ValueError("Invalid setting {response} for response parameter, must be one of {'warn', 'error'}")


def _detect_nan_inf(estimates, response="warn"):
    """Detect whether array of parameter estimates contains NaN/inf."""
    if not (np.isfinite(estimates)).all():
        if response=="error":
            raise ValueError("Zero or negative values detected")
        elif response=="warn":
            warnings.warn("NaNs or Infs detected in input - information criterion cannot be calculated", UserWarning)
        else:
             raise ValueError("Invalid setting {response} for response parameter, must be one of {'warn', 'error'}")


def claic(f, estimates, step=1e-2, step_factor=0.5, zeros="allow", nan_infs="error"):
    """Compute composite likelihood Akaike information criterion (CLAIC).

    Args:
        f (function): Composite log-likelihood function.
        estimates (ndarray): Parameter estimates at which to evaluate CLAIC.
        step (float): Step size for finding Hessian and Jacobian matrices. Defaults to 1e-2.
        step_factor (float): Factor by which to multiply the step size in each attempt. Defaults to 0.5.
        zeros (str, optional): What to do with zeros in input, one of {"allow", "warn", "error"}. Defaults to "allow".
        nan_infs (str, optional): What to do with NaNs and infs in input, one of {"allow", "warn", "error"}. Defaults to "error".

    Returns:
        float: Composite likelihood AIC estimate.
    """
    if nan_infs != "allow":
        _detect_nan_inf(estimates, response=nan_infs)
    if zeros != "allow":
        _detect_zeros(estimates, response=zeros)

    likelihood = f(estimates)
    jacobian_mat, hessian_mat = _make_matrices(f, estimates, step, step_factor)
    return _cl_akaike(likelihood, jacobian_mat, hessian_mat)


def clbic(f, estimates, n_samples, step=1e-2, step_factor=0.5, zeros="allow", nan_infs="error"):
    """Compute composite likelihood Bayesian information criterion (CLBIC).

    Args:
        f (function): Composite log-likelihood function.
        estimates (ndarray): Parameter estimates at which to evaluate CLBIC.
        n_samples (int): Number of samples in data.
        step (float): Step size for finding Hessian and Jacobian matrices. Defaults to 1e-2.
        step_factor (float): Factor by which to multiply the step size in each attempt. Defaults to 0.5.
        zeros (str, optional): What to do with zeros in input, one of {"allow", "warn", "error"}. Defaults to "allow".
        nan_infs (str, optional): What to do with NaNs and infs in input, one of {"allow", "warn", "error"}. Defaults to "error".

    Returns:
        float: Composite likelihood BIC estimate.
    """
    if nan_infs != "allow":
        _detect_nan_inf(estimates, response=nan_infs)
    if zeros != "allow":
        _detect_zeros(estimates, response=zeros)

    likelihood = f(estimates)
    jacobian_mat, hessian_mat = _make_matrices(f, estimates, step, step_factor)
    return _cl_bayes(likelihood, jacobian_mat, hessian_mat, n_samples)














